import argparse
import glob
import os
import re
from xml.dom import minidom
from threading import Thread
import time
from src.classes.vivadoBatchRun import vivadoBatchRun
import sys
import xml.etree.ElementTree as ET

if __name__ == "__main__":

    vivadoBatchLoc = os.path.split(os.path.realpath(__file__))[0]
    parser = argparse.ArgumentParser(description='Minerva command line options.')

    parser.add_argument('-tp', '--Minerva_TP_Opt',
        help='Runs Minerva in Minerva_TP_Opt mode to maximize the frequency',
        action='store_true')
    parser.add_argument('-tpa', '--Minerva_TPA_Opt',
        help='Runs Minerva in Minerva_TPA_Opt mode to maximize the \
        frequency/LUT ratio', action='store_true')
    parser.add_argument('-fast', '--Minerva_Fast_Opt',
        help='Runs Minerva in Minerva_Fast_Opt mode. It only runs optimization \
        number 11', action='store_true')
    parser.add_argument('-sr', '--single_run',
        help='Runs single vivado run for specific frequency and opt id',
        action='store_true')
    parser.add_argument('--short_report',
        help='Specify the design name to print out its short report',
        action='store_true')
    parser.add_argument('-respkg', '--gen_resrep_pkg',
        help='Generates result replication package for the user specified \
        design and result', action='store_true')
    parser.add_argument('-chpart', '--change_part',
        help='Change Part ID based on the device list available in \
        config/parts.xml', action='store_true')
        
    parser.add_argument('--req_freq',
        help='Specify the requested frequency required for single run',
        default=150)
    parser.add_argument('--opt_id',
        help='Specify the optimization id required for single run',
        default=2)
    parser.add_argument('--run_id',
        help='Specify the run id based on the runs specified in runs.xml \
        (this argument is only available in single_run)',
        default=0)
    parser.add_argument('--part_id',
        help='Specify Part ID based on the device list available in config/parts.xml',
        default=0)
    parser.add_argument('--init_l_f',
        help='Specify frequency range start value (lower bound) in MHz',
        default=100)
    parser.add_argument('--init_u_f',
        help='Specify frequency range end value (upper bound) in MHz',
        default=500)
    parser.add_argument('-r', '--range_size',
        help='Specify range size',
        default=12)
    parser.add_argument('-n', '--num_runs',
        help='Specify max number of runs in parallel',
        default=6)
    parser.add_argument('-p', '--num_opts',
        help='Specify number of optimization strategies',
        default=23)
    parser.add_argument('-fo', '--fast_opt_ids',
        help='Specify optimization strategies to be used for Minerva_Fast_Opt mode',
        default=[11], nargs='*')

    parser.add_argument('-rir', '--run_id_range',
        help='Specify custom run id range corresponding to runs.xml config file to run. This \
        argument receives exactly 2 integers (lower bound and upper bound) as input.',
        nargs=2)

    parser.add_argument('-gr', '--gen_runs',
        help='Specify workspace location to generates runs.xml config file automatically')

    parser.add_argument('-top', '--top_name',
        help='Specify top-level module name',
        default= -1)

    parser.add_argument('--debug', help='Debugging purposes',
        action='store_true')

    args = parser.parse_args()

    SrcLoc = str(args.gen_runs)
    InitLF = int(args.init_l_f)
    InitUF = int(args.init_u_f)
    RangeSize = int(args.range_size)
    MaxNumRuns = int(args.num_runs)
    PartId = int(args.part_id)
    ReqF = int(args.req_freq)
    OptId = int(args.opt_id)
    RunId = int(args.run_id)
    NumOpts = int(args.num_opts)
    topName = args.top_name

    def prettify(elem):
        """Return a pretty-printed XML string for the Element.
        """
        rough_string = ET.tostring(elem, 'utf-8')
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="\t")

    if args.gen_runs:
        design_list_path = []
        design_list_name = []
        if os.path.exists(SrcLoc):
            print "Workspace directory:\n" + SrcLoc
            print "The following designs successfully added to the runs.xml:"
        else:
            sys.exit('ERROR: specified workspace directory(' + SrcLoc + ') doese not exists')

        design_list = glob.glob(SrcLoc+ '/*')
        for i in range(len(design_list)):
            if os.path.exists(os.path.abspath(design_list[i] + '/src_rtl/')):
                if os.path.exists(design_list[i] + '/src_rtl/source_list.txt'):
                    design_list_path.append(design_list[i])
                    design_name = design_list[i].replace(str(SrcLoc), '')
                    design_name = design_name.replace('\\', '')
                    design_list_name.append(design_name)
                    print design_list_name[i]
                else:
                    print ('ERROR: ' + design_list[i] + ' does not contain source_list.txt')
            else:
                print ('ERROR: ' + design_list[i] + ' does not contain src_rtl directory')

        vbr_elm = ET.Element('vivadoBatchRun', attrib={'isOpt': '0', 'isTD': '1',
            'numRuns': str(len(design_list_path)), 'partIdx': str(PartId)})
        vrl_elm = ET.SubElement(vbr_elm, 'vivadoRunsList')
        for i in range(len(design_list_path)):
            run_elm = ET.SubElement(vrl_elm, 'vivadoRun', attrib={'id': str(i)})
            loc_elm = ET.SubElement(run_elm, 'location', attrib={'algName': design_list_name[i]})
            loc_elm.text = design_list_path[i]
        with open (os.path.abspath(vivadoBatchLoc +
            "/config/runs.xml"), 'w') as runs_out:
            runs_out.write(prettify(vbr_elm))

    if args.part_id:
        tree = ET.parse(os.path.abspath(vivadoBatchLoc + "/config/runs.xml"))
        root = tree.getroot()
        root.set('partIdx', str(PartId))
        tree.write(os.path.abspath(vivadoBatchLoc + "/config/runs.xml"))

    if args.change_part:
        runs_xml_path = os.path.abspath(vivadoBatchLoc + "/config/runs.xml")
        part_xml_path = os.path.abspath(vivadoBatchLoc + "/config/parts.xml")
        id_list = []
        print "\nAvailable devices in parts.xml:"
        tree = ET.parse(part_xml_path)
        root = tree.getroot()
        for part in root.iter('part'):
            print ("Part ID " + part.get('id') + ": (" + part.get('family') + 
                ")  " + part.text) 
        userIn_part_id = input('\nEnter part id to be used for rusult generation: ')
        #TODO: Check for the correct input        
        tree = ET.parse(runs_xml_path)
        root = tree.getroot()
        root.set('partIdx', str(userIn_part_id))
        tree.write(runs_xml_path)            
        
    if args.run_id_range:
        run_start_id = int(args.run_id_range[0])
        run_end_id = int(args.run_id_range[1])
        if run_start_id > run_end_id :
            sys.exit("ERROR: Incorrect id range")

    else:
        run_start_id = -1
        run_end_id = -1

    vBR = vivadoBatchRun(vivadoBatchLoc, topName)

    if args.single_run:
        print "single run"
        vBR.SingleRun(ReqF, OptId, RunId)

    if args.Minerva_TP_Opt:
		vBR.Minerva_TP_Opt(InitLF, InitUF, RangeSize, MaxNumRuns, NumOpts,
        run_start_id, run_end_id)

    if args.Minerva_TPA_Opt:
		vBR.Minerva_TPA_Opt(InitLF, InitUF, RangeSize, MaxNumRuns, NumOpts, run_start_id,
        run_end_id)

    if args.Minerva_Fast_Opt:
        fast_num_opts = len(args.fast_opt_ids)+1
        vBR.Minerva_Fast_Opt(InitLF, InitUF, RangeSize, MaxNumRuns, fast_num_opts,
        args.fast_opt_ids, run_start_id, run_end_id)

    if args.short_report:
        run_xml_path = os.path.abspath(vivadoBatchLoc + "/config/runs.xml")
        print "\nAvailable designs in run.xml:"
        tree = ET.parse(run_xml_path)
        root = tree.getroot()
        for run in root.iter('vivadoRun'):
            run_elm = run.find('location')
            print ("Run ID " + run.get('id') + ": " + run_elm.get('algName'))
        userIn_design_id = input('\nEnter design id to see the short_report: ')
        for run in root.iter('vivadoRun'):
            if int(run.get('id'))== userIn_design_id:
                run_elm = run.find('location')
                alg_name = run_elm.get('algName')
                alg_loc = run_elm.text
                run_exist = 1
        if not run_exist:
            sys.exit("ERROR: Incorrect run id entered")

        print "Available results for "+ alg_name +":\n"
        status_xml_path = os.path.abspath(alg_loc + "/minerva_status/" +
            alg_name + "_MS.xml")
        if not os.path.exists(status_xml_path):
            sys.exit('ERROR: No result available for ' + alg_name)

        tree = ET.parse(status_xml_path)
        ms_elm = tree.getroot()
        sl_elm = ms_elm.find('StatusList')
        for result_elm in sl_elm:
            for st_elm in result_elm:
                status_elm = st_elm
            if int(result_elm.get('is_complete')) == 1:
                print ("Result ID " + result_elm.get('id') + " (Complete): " +
                    ET.tostring(status_elm, 'utf-8'))
            else:
                print ("Result ID " + result_elm.get('id') + " (In progress or terminated): " +
                    ET.tostring(status_elm, 'utf-8'))

        userIn_result_id = input('Enter result id to see the short short: ')
        report_name = alg_name + "_ID" + str(userIn_result_id)
        vBR.ShortReportGen(report_name, alg_loc)


    if args.gen_resrep_pkg:
        run_exist = 0
        result_exist = 0

        run_xml_path = os.path.abspath(vivadoBatchLoc + "/config/runs.xml")
        print "\nAvailable designs in run.xml:"
        tree = ET.parse(run_xml_path)
        root = tree.getroot()
        for run in root.iter('vivadoRun'):
            run_elm = run.find('location')
            print ("Run ID " + run.get('id') + ": " + run_elm.get('algName'))

        userIn_design_id = input('\nEnter design id for result replication package: ')

        for run in root.iter('vivadoRun'):
            if int(run.get('id'))== userIn_design_id:
                run_elm = run.find('location')
                alg_name = run_elm.get('algName')
                alg_loc = run_elm.text
                run_exist = 1
        if not run_exist:
            sys.exit("ERROR: Incorrect run id entered")

        print "Available results for "+ alg_name +":\n"
        status_xml_path = os.path.abspath(alg_loc + "/minerva_status/" +
            alg_name + "_MS.xml")
        if not os.path.exists(status_xml_path):
            sys.exit('ERROR: No result available for ' + alg_name)

        tree = ET.parse(status_xml_path)
        ms_elm = tree.getroot()
        sl_elm = ms_elm.find('StatusList')
        for result_elm in sl_elm:
            for st_elm in result_elm:
                status_elm = st_elm
            if int(result_elm.get('is_complete')) == 1:
                print ("Result ID " + result_elm.get('id') + " (Complete): " +
                    ET.tostring(status_elm, 'utf-8'))
            else:
                print ("Result ID " + result_elm.get('id') + " (In progress or terminated): " +
                    ET.tostring(status_elm, 'utf-8'))

        userIn_result_id = input('Enter result id for result replication package generation: \n')

        for result_elm in sl_elm:
            if int(result_elm.get('id')) == userIn_result_id:
                for st_elm in result_elm:
                    status_elm = st_elm
                freq = float(status_elm.get('freq'))
                opt_id = int(status_elm.get('opt_id'))
                top_str = status_elm.get('top_name')
                tool_ver = (status_elm.get('tool_version'))[0:6]
                result_exist = 1
        if not result_exist:
            sys.exit("ERROR: Incorrect result id entered")

        userIn_src_mode = input('Enter source code inclusion mode: \n' +
            '\n1- Include all source code in the package' +
            '\n2- Include source code url in the package' +
            '\n3- Do not include source code in the package \n')
        userIn_url_text = ""
        if int(userIn_src_mode) == 2:
            userIn_url_text = raw_input('Enter source code url: ')

        vBR.ResultReplicationPkgGen(freq, opt_id, userIn_design_id, top_str,
            tool_ver, userIn_result_id, int(userIn_src_mode), userIn_url_text)

    if args.debug:
        vBR.debug()
