import os
import glob
from xml.dom import minidom
from xml.dom import Node
import xml.etree.ElementTree as ET
import time
from threading import Thread
from src.classes.vivadoConfig import vivadoConfig

class Borg:
    _shared_state = {}
    def __init__(self):
        self.__dict__ = self._shared_state

class vivadoBatchRun(Borg):
    XDCFILEPREFIX = "freq"
    isInit = 0

    def __init__(self, vivadoBatchLoc, topName):
        """ Create a new Vivado Run configuration object """
        Borg.__init__(self)

        self.vivadoConf = vivadoConfig()
        vivadoBatchRun.vivadoBatchLoc = vivadoBatchLoc
        self.genFilename = ""
        self.runFilename = ""
        self.run25Filename = ""

        self.projModeGenFilename = ""
        self.projModeOptGenFilename = ""
        self.projModeBitGenFilename = ""

        self.top_name = topName

        if not vivadoBatchRun.isInit:
            vivadoBatchRun.isInit = 1
            self.runInit()

        if not os.path.exists(vivadoBatchLoc + "/reports"):
            os.makedirs(vivadoBatchLoc + "/reports")

        #TODO move it to tcl file generation function
        #! Generate Tcl files based on template Tcl scripts and current location
        tcl_home_path = vivadoBatchLoc + "/scripts/Tcl"
        tcl_home_path = tcl_home_path.replace('\\', '/')
        tcl_list = ["genTD", "runTD", "runTD_25", "runTD_ResRep"]
        for i in range (len(tcl_list)):
            with open (os.path.abspath(vivadoBatchLoc + "/scripts/" +
                tcl_list[i] + "_temp.tcl"), 'r') as tcl_file_in:
                with open (os.path.abspath(vivadoBatchLoc + "/scripts/" +
                    tcl_list[i] + ".tcl"), 'w') as tcl_file_out:
                    tcl_file_data = tcl_file_in.read()
                    tcl_file_out.write(tcl_file_data.replace("set tclHome",
                        ("set tclHome \"" + tcl_home_path +"\"")))

    def parseXML(self):

        runsConfigLoc = os.path.abspath(vivadoBatchRun.vivadoBatchLoc + "/config/runs.xml")
        partsConfigLoc = os.path.abspath(vivadoBatchRun.vivadoBatchLoc + "/config/parts.xml")

        #populate partsList
        partsList = []
        tree = ET.parse(partsConfigLoc)
        root = tree.getroot()
        for part in root.iter('part'):
            partsList.append(part.text)
        
        #populate vivadoRunsList
        self.vivadoRunsList = []
        from vivadoRun import vivadoRun
        tree = ET.parse(runsConfigLoc)
        root = tree.getroot()
        self.numRuns = int(root.get('numRuns'))
        self.isTD = int(root.get('isTD'))
        self.isOpt = int(root.get('isOpt'))
        self.partIdx = int(root.get('partIdx'))
        for run in root.iter('location'):
            alg_name = run.get('algName')
            loc = os.path.abspath(run.text)
            self.vivadoRunsList.append(vivadoRun(loc, alg_name, self.top_name,
                partsList[int(self.partIdx)], self))

    def runInit(self):
        self.parseXML()

        if (self.isTD):
            self.genFilename = "genTD"
            self.runFilename = "runTD"
            self.run25Filename = "runTD_25"

            self.projModeGenFilename = "projmode_testbed"
            self.projModeOptGenFilename = "projmode_testbed_opt"
            self.projModeBitGenFilename = "Pmode_testbed_bitGen"
        else:
            self.genFilename = "genOOC"
            self.runFilename = "runOOC"
            self.run25Filename = "runOOC_25"

    def ResultReplicationPkgGen(self, ReqF, OptId, RunId, TopStr, ToolVer,
        ResId, SrcMode, UrlText):
        optString = self.vivadoConf.getStratString(ReqF, OptId)
        self.vivadoRunsList[RunId].ResultReplicationPkgGen(ReqF, OptId,
            optString, TopStr, ToolVer, ResId, SrcMode, UrlText)

    def SingleRun(self, ReqF, OptId, RunId):

        threadList = []
        optString = self.vivadoConf.getStratString(ReqF, OptId)
        threadList.append(Thread(target=self.vivadoRunsList[RunId].SingleRun,
            args=(ReqF, OptId, optString)))
        time.sleep(2)
        threadList[0].start()
        time.sleep(2)
        threadList[0].join()

    def Minerva_TP_Opt(self, InitLF, InitUF, RangeSize, MaxNumRuns, NumOpts,
        StartID, EndID):
        alg_name = "Minerva_TP_Opt"
        print "\n========== Minerva_TP_Opt Routine ==========\n"
        if self.numRuns > len(self.vivadoRunsList):
            print "Error in config file, size numRuns is greater than the \
            number of vivadoRun entries.  Please fix to continue"
            return
        elif EndID > self.numRuns:
            print "ERROR: run_end_id is higher than number of available runs in \
            runs.xml config file"
        start_id = StartID if StartID >=0 else 0
        end_id = EndID if EndID >=0 else (self.numRuns -1)
        num_runs = (end_id - start_id + 1) if StartID >=0 else self.numRuns
        self.MaxFreqList = [0 for j in range(self.numRuns)]
        threadList = []
        optList_2D = [[0 for x in range(NumOpts+1)] for y in range(RangeSize)]
        self.MaxFreqAreaList = [0 for j in range(self.numRuns)]
        self.BestAreaList = [0 for j in range(self.numRuns)]
        for i in range(RangeSize):
            for j in range(1,NumOpts+1):
                optList_2D[i][j]=self.vivadoConf.getStratString(i,j)

        for j in range(start_id, end_id+1):
            threadList.append(Thread(target=self.vivadoRunsList[j].Minerva_TP_Opt,
            args=(InitLF, InitUF, RangeSize, MaxNumRuns, self.MaxFreqList, j,
                NumOpts, optList_2D, alg_name)))

        for j in range(num_runs):
            time.sleep(10)
            threadList[j].start()
            threadList[j].join()

        print self.MaxFreqList

    def Minerva_TPA_Opt(self, InitLF, InitUF, RangeSize, MaxNumRuns, NumOpts,
        StartID, EndID):
        alg_name = "Minerva_TPA_Opt"
        print "\n========== Minerva_TPA_Opt Routine ==========\n"
        if self.numRuns > len(self.vivadoRunsList):
            print "Error in config file, size numRuns is greater than the \
            number of vivadoRun entries.  Please fix to continue"
            return
        elif EndID > self.numRuns:
            print "ERROR: run_end_id is higher than number of available runs in \
            runs.xml config file"
        start_id = StartID if StartID >=0 else 0
        end_id = EndID if EndID >=0 else (self.numRuns -1)
        num_runs = (end_id - start_id + 1) if StartID >=0 else self.numRuns
        self.MaxFreqList = [0 for j in range(self.numRuns)]
        threadList = []
        optList_2D = [[0 for x in range(NumOpts+1)] for y in range(100)]
        self.MaxFreqAreaList = [0 for j in range(self.numRuns)]
        self.BestAreaList = [0 for j in range(self.numRuns)]
        for i in range(100):
            for j in range(1,NumOpts+1):
                optList_2D[i][j]=self.vivadoConf.getStratString(i,j)

        for j in range(start_id, end_id+1):
            threadList.append(Thread(target=self.vivadoRunsList[j].Minerva_TP_Opt,
            args=(InitLF, InitUF, RangeSize, MaxNumRuns, self.MaxFreqList, j,
                NumOpts, optList_2D, alg_name)))

        for j in range(num_runs):
            time.sleep(10)
            threadList[j].start()
            threadList[j].join()

        print self.MaxFreqList
        del threadList [:]

        for j in range(start_id, end_id+1):
            threadList.append(Thread(target=self.vivadoRunsList[j].Minerva_TPA_Opt,
                args=(self.MaxFreqList[j], (NumOpts+1), optList_2D, MaxNumRuns,
                self.MaxFreqAreaList, self.BestAreaList, j)))

        for j in range(num_runs):
            time.sleep(10)
            threadList[j].start()
            threadList[j].join()

        print self.MaxFreqAreaList
        print self.BestAreaList

    def Minerva_Fast_Opt(self, InitLF, InitUF, RangeSize, MaxNumRuns, NumOpts,
        FastOptIdList, StartID, EndID):
        alg_name = "Minerva_Fast_Opt"
        print "\n========== Minerva_Fast_Opt Routine ==========\n"
        if self.numRuns > len(self.vivadoRunsList):
            print "Error in config file, size numRuns is greater than the \
            number of vivadoRun entries.  Please fix to continue"
            return
        elif EndID > self.numRuns:
            print "ERROR: run_end_id is higher than number of available runs in \
            runs.xml config file"
        start_id = StartID if StartID >=0 else 0
        end_id = EndID if EndID >=0 else (self.numRuns -1)
        num_runs = (end_id - start_id + 1) if StartID >=0 else self.numRuns
        self.MaxFreqList = [0 for j in range(self.numRuns)]
        threadList = []
        optList_2D = [[0 for x in range(NumOpts+1)] for y in range(RangeSize)]
        self.MaxFreqAreaList = [0 for j in range(self.numRuns)]
        self.BestAreaList = [0 for j in range(self.numRuns)]
        for i in range(RangeSize):
            for j in range(1,NumOpts+1):
                if j==1:
                    optList_2D[i][j]=self.vivadoConf.getStratString(i,j)
                else:
                    optList_2D[i][j]=self.vivadoConf.getStratString(i,int(FastOptIdList[j-2]))

        for j in range(start_id, end_id+1):
            threadList.append(Thread(target=self.vivadoRunsList[j].Minerva_TP_Opt,
            args=(InitLF, InitUF, RangeSize, MaxNumRuns, self.MaxFreqList, j,
                NumOpts, optList_2D, alg_name)))

        for j in range(num_runs):
            time.sleep(10)
            threadList[j].start()
            threadList[j].join()

        print self.MaxFreqList

    def TPtoAreaOptSearch(self, stratID):
        if self.numRuns > len(self.vivadoRunsList):
            print "Error in config file, size numRuns is greater than the number of vivadoRun entries.  Please fix to continue"
            return

        self.MaxFreqAreaList = [0 for j in range(self.numRuns)]
        self.BestAreaList = [0 for j in range(self.numRuns)]
        threadList = []
        MaxFreqList = [347, 296]

        for j in range(self.numRuns):

            threadList.append(Thread(target=self.vivadoRunsList[j].Minerva_TPA_Opt,
                args=(MaxFreqList[j], 24, 16, self.MaxFreqAreaList,
                self.BestAreaList, j)))

        for j in range(self.numRuns):
            time.sleep(10)
            threadList[j].start()
            threadList[j].join()

    def runAllBinaryFreqSearch(self, ID):

        if self.numRuns > len(self.vivadoRunsList):
            print "Error in config file, size numRuns is greater than the number of vivadoRun entries.  Please fix to continue"
            return

        self.MaxFreqList = [0 for j in range(self.numRuns)]
        threadList = []

        for j in range(self.numRuns):
            threadList.append(Thread(target=self.vivadoRunsList[j].BinaryFreqSearch,
                args=(50, 500, j, self.MaxFreqList)))

        for j in range(self.numRuns):
            time.sleep(10)
            threadList[j].start()
        time.sleep(10)
        for j in range(self.numRuns): threadList[j].join()

        print self.MaxFreqList

    def ShortReportGen(self, ReportName, ReportLoc):
        print "\n\n ************* SHORT REPORTS ************* \n\n"
        read_flag = 0
        report_loc = os.path.abspath(ReportLoc +
            "/minerva_status/full_reports/" + ReportName + ".txt")
        with open (report_loc, 'r') as rep_in:
            for line in rep_in:
                if "**** Minerva_TP_Opt DONE ****" in line:
                    result_exist = 1
                    read_flag = 1
                    print "\n______________ " + ReportName + " ___________\n\n"
                elif "**** Minerva_TPA_Opt DONE ****" in line:
                    result_exist = 1
                    read_flag = 1
                    print "\n______________ " + ReportName + " ___________\n\n"
                elif "**** Minerva_Fast_Opt DONE ****" in line:
                    result_exist = 1
                    read_flag = 1
                    print "\n______________ " + ReportName + " ___________\n\n"
                elif "**** Single Run DONE ****" in line:
                    result_exist = 1
                    read_flag = 1
                    print "\n______________ " + ReportName + " ___________\n\n"
                elif (read_flag):
                    print line
                    if "Report Summary End" in line:
                        read_flag = 0
            if not result_exist:
                print "\n\n ERROR: No results found in " + report_list[i] + ". \n\n"

    def debug(self):
        self.vivadoRunsList[0].debug()
