import os
import sys
import time
import subprocess
import re
from threading import Thread
from vivadoBatchRun import vivadoBatchRun
import shutil
import datetime
from vivadoConfig import vivadoConfig
import xml.etree.ElementTree as ET
from xml.dom import minidom

class vivadoRun(object):
    """ Create a Vivado Run object, that will use vivado batch mode to generate
        an optimized result for a given circuit
    """

    def __init__(self, prjLoc, designName, topName, partStrList, vBR):
        """ Create a new Vivado Run object"""
        """ prjLoc - Location of project containing src_rtl folder"""

        if not vivadoBatchRun.isInit:
            print "Error vivadoBatchRun object is not initialized, no run created."
            sys.stdout.flush()
            return
        else: self.vBR = vBR

        self.prjLoc = prjLoc
        self.partStrList = partStrList
        self.topStr = "none"
        self.design_name = designName
        self.top_name = topName
        self.report_loc = os.path.abspath(self.prjLoc +
            "/minerva_status/full_reports/")
        if not os.path.exists(os.path.abspath(self.prjLoc + "/runs/")):
            os.makedirs(os.path.abspath(self.prjLoc + "/runs/"))

        self.report_num = 1
        while os.path.exists(os.path.abspath(self.vBR.vivadoBatchLoc +
            "/reports/" + self.design_name + "_" + str(self.report_num) +
            ".out")):
            self.report_num +=1

        if not os.path.exists(os.path.abspath(self.prjLoc + "/minerva_status/" +
            self.design_name + "_MS.xml")):
            os.makedirs(os.path.abspath(self.prjLoc + "/minerva_status/"))
            self.ms_elm = ET.Element('MinervaStatus', attrib={'AlgName':
                self.design_name,'NumStatus': '0'})
            sl_elm = ET.SubElement(self.ms_elm, 'StatusList')
            with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                + self.design_name + "_MS.xml"), 'w') as status_xml:
                status_xml.write(ET.tostring(self.ms_elm, 'utf-8'))
        else:
            tree = ET.parse(os.path.abspath(self.prjLoc +"/minerva_status/" +
                self.design_name + "_MS.xml"))
            self.ms_elm = tree.getroot()

        if not os.path.exists(os.path.abspath(self.report_loc)):
            os.makedirs(os.path.abspath(self.report_loc))

    def prettify(self, elem):
        """Return a pretty-printed XML string for the Element.
        """
        rough_string = ET.tostring(elem, 'utf-8')
        reparsed = minidom.parseString(rough_string)
        return '\n'.join([line for line in reparsed.toprettyxml(indent=' '*4).
            split('\n') if line.strip()])

    def prettify_2(self, elem):
        """Return a pretty-printed XML string for the Element.
        """
        rough_string = ET.tostring(elem, 'utf-8')
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="\t")

    def statusFileGen(self, mode):
        timestr = time.strftime("%Y-%m-%d %H:%M")
        num_status = self.ms_elm.get('NumStatus')
        new_num_status = str(int(num_status)+1)
        self.ms_elm.set('NumStatus', new_num_status)
        sl_elm = self.ms_elm.find('StatusList')
        result_elm = ET.SubElement(sl_elm, 'MinervaResult', attrib={
            'id': str(num_status), 'is_complete': '0'})
        run_elm = ET.SubElement(result_elm, str(mode), attrib={
            'FF': '0', 'Slice': '0', 'freq': '0', 'LUT': '0', 'run_time': '0',
            'opt_id': '0', 'start_time': str(timestr), 'top_name': self.topStr,
            'tool_version': '0', 'device': self.partStrList})
        with open (os.path.abspath(self.prjLoc + "/minerva_status/"
            + self.design_name + "_MS.xml"), 'w') as status_xml:
            status_xml.write(self.prettify(self.ms_elm))
        return result_elm

    def getOptID(self, OptStr):
        opt_id = re.findall("\d+",OptStr)
        return int(opt_id[0])

    def getVersion(self):
        WNSSeekCnt = 0
        with open (os.path.abspath(self.prjLoc + "/runs/"+self.partStrList+
            "/Synth/freq0/" + self.topStr + "/" + self.topStr +
            "_utilization_synth.rpt"), 'r') as fin:
            for line in fin:
                if re.match("(.*)Tool Version(.*)", line) or WNSSeekCnt:
                    v_str = line[26:40]
                    return v_str

    def getWNS(self, f_idx, o_idx):
        WNSSeekCnt = 0
        freqSel = "freq" + str(f_idx)
        TopDownsel = "TopDown" + str(o_idx)
        with open (os.path.abspath(self.prjLoc + "/runs/"+self.partStrList+
            "/Implement/" + freqSel + "/" + TopDownsel + "/reports/" +
            self.topStr + "_timing_summary.rpt"), 'r') as fin:
            for line in fin:
                if re.match("(.*)Design Timing Summary(.*)", line) or WNSSeekCnt:
                    if not WNSSeekCnt:
                        WNSSeekCnt = 6
                    else:
                        WNSSeekCnt = WNSSeekCnt - 1
                        if not WNSSeekCnt:
                            valList = re.findall("-?\d+\.\d+", line)
                            return float(valList[0])

    def getLUT(self, f_idx, o_idx):
        freqSel = "freq" + str(f_idx)
        TopDownsel = "TopDown" + str(o_idx)
        with open (os.path.abspath(self.prjLoc + "/runs/"+self.partStrList+
            "/Implement/" + freqSel + "/" + TopDownsel + "/reports/" +
            self.topStr + "_utilization_route_design.rpt"), 'r') as fin:
            for line in fin:
                if re.match("(.*)Slice LUTs(.*)", line):
                    valList = re.findall("\d+", line)
                    return int(valList[0])
                elif re.match("(.*)CLB LUTs(.*)", line):
                    valList = re.findall("\d+", line)
                    return int(valList[0])    

    def getSlice(self, f_idx, o_idx):
        freqSel = "freq" + str(f_idx)
        TopDownsel = "TopDown" + str(o_idx)
        with open (os.path.abspath(self.prjLoc + "/runs/"+self.partStrList+
            "/Implement/" + freqSel + "/" + TopDownsel + "/reports/" +
            self.topStr + "_utilization_route_design.rpt"), 'r') as fin:
            for line in fin:
                if re.match("(.*)Slice        (.*)", line):
                    valList = re.findall("\d+", line)
                    return int(valList[0])
                elif re.match("(.*)CLB        (.*)", line):
                    valList = re.findall("\d+", line)
                    return int(valList[0])

    def getFF(self, f_idx, o_idx):
        freqSel = "freq" + str(f_idx)
        TopDownsel = "TopDown" + str(o_idx)
        with open (os.path.abspath(self.prjLoc + "/runs/"+self.partStrList+
            "/Implement/" + freqSel + "/" + TopDownsel + "/reports/" +
            self.topStr + "_utilization_route_design.rpt"), 'r') as fin:
            for line in fin:
                if re.match("(.*)Register as Flip Flop(.*)", line):
                    valList = re.findall("\d+", line)
                    return int(valList[0])

    def runGen(self, f_idx):
        freqSel = "freq" + str(f_idx)
        os.chdir(self.prjLoc)
        #p = subprocess.Popen("cd " + self.prjLoc +
        #    "; vivado -mode batch -source " + self.vBR.genFilename + "_" +
        #    freqSel + ".tcl -notrace > synth_out" + str(f_idx) +
        #    ".txt", shell=True)
        p = subprocess.Popen(
            "vivado -mode batch -source " + self.vBR.genFilename + "_" +
            freqSel + ".tcl -notrace > synth_out" + str(f_idx) +
            ".txt", shell=True)

        while p.poll() is None:
            time.sleep(30)

        os.chdir(self.prjLoc)
        #p = subprocess.Popen("cd " + self.prjLoc +
        #    "; vivado -mode batch -source " + self.vBR.runFilename + "_" +
        #    freqSel + ".tcl -notrace > impl_out" + str(f_idx) +
        #    ".txt", shell=True)

        p = subprocess.Popen(
            "vivado -mode batch -source " + self.vBR.runFilename + "_" +
            freqSel + ".tcl -notrace > impl_out" + str(f_idx) +
            ".txt", shell=True)
        while p.poll() is None:
            time.sleep(30)

    def run(self, f_idx, o_idx, optString):
        freqSel = "freq" + str(f_idx)
        stratDefFound = 0
        fileInName = os.path.abspath(self.prjLoc + "/" + self.vBR.runFilename +
            "_" + freqSel + ".tcl")
        fileOutName = os.path.abspath(self.prjLoc + "/" + self.vBR.runFilename +
            "_" + freqSel + "_" + str(o_idx) + ".tcl")

        #create gen/run TCL files for execution of runs
        with open(fileInName, 'r') as genFileIn:
            with open(fileOutName, 'w') as genFileOut:
                for line in genFileIn:
                    if not stratDefFound:
                        genFileOut.write(line)
                    if "stratDef" in line:
                            if not stratDefFound:
                                stratDefFound = 1
                            else: #last stratDef, write strat
                                genFileOut.write(optString)
                                genFileOut.write(line)
                                stratDefFound = 0

        os.chdir(self.prjLoc)
        #p = subprocess.Popen("cd " + self.prjLoc +
        #    "; vivado -mode batch -source " + fileOutName +
        #    " -notrace > impl_out" + str(f_idx) + ".txt", shell=True)

        p = subprocess.Popen(
            "vivado -mode batch -source " + fileOutName +
            " -notrace > impl_out" + str(f_idx) + ".txt", shell=True)
        while p.poll() is None:
            time.sleep(30)

        os.remove(os.path.abspath(fileOutName))

    def createXDCClock(self, freqFileName, clockVal):
        with open (os.path.abspath(freqFileName), 'w') as f:
            f.write("create_clock -period " + "{0:.3f}".format(clockVal) +
                " -name clk -waveform {0.000 " +
                "{0:.3f}".format(float(clockVal/2)) + "} [get_ports clk]");
        f.close();

    def createClockFiles(self, n, f0, fn, freqList, periodList):
		diff = float(fn - f0)
		step = (diff)/n
		for i in range(n):
			currFreq = float(f0+(i*step))
			currPeriod = float(1000/currFreq)
			periodList.append(currPeriod)
			freqList.append(currFreq)
			self.createXDCClock(self.prjLoc + "/" +
                vivadoBatchRun.XDCFILEPREFIX + str(i) + ".xdc", currPeriod)

    def createClockFiles_1MhzStep(self, f0, fn, freqList, periodList):
        j = 0
        for i in range(f0, (fn+1)):
            currFreq = float(i*1)
            currPeriod = float(1000/currFreq)
            sys.stdout.flush()
            periodList.append(currPeriod)
            freqList.append(i)
            self.createXDCClock(self.prjLoc + "/" +
                vivadoBatchRun.XDCFILEPREFIX + str(j) + ".xdc", currPeriod)
            j +=1

    def createTCLFile(self, fileName, refInputData, freqID):
        newGenFileName = fileName + "_freq" + str(freqID) + ".tcl"
        refInputData = refInputData.replace("./synthDir","./runs/" +
            self.partStrList + "/Synth/freq" + str(freqID))
        refInputData = refInputData.replace("./implDir","./runs/" +
            self.partStrList + "/Implement/freq" + str(freqID))
        refInputData = refInputData.replace("./dcpDir","./runs/" +
            self.partStrList + "/Checkpoint/freq" + str(freqID))
        refInputData = refInputData.replace("${top}_flpn.xdc", "freq" +
            str(freqID) + ".xdc")
        refInputData = refInputData.replace("set top \"uut_top\"", "set top \"" +
            self.topStr + "\"")
        refInputData = refInputData.replace("device", self.partStrList)
        #refInputData = refInputData.replace("devicePKG", self.partStrList[1])
        #refInputData = refInputData.replace("deviceSPEED", self.partStrList[2])
        #refInputData = refInputData.replace("deviceTEMP", self.partStrList[3])

        with open(os.path.abspath(self.prjLoc + "/" +
            newGenFileName), 'w') as genFileOut:
            genFileOut.write(refInputData)

    def createTCLFile_ResRep(self, fileName, refInputData, freqID, optID,
        optString):
        stratDefFound = 0
        newGenFileName = fileName + "_freq" + str(freqID) + "_" + str(optID) + ".tcl"
        refInputData = refInputData.replace("./synthDir","./runs/" +
            self.partStrList + "/Synth/freq" + str(freqID))
        refInputData = refInputData.replace("./implDir","./runs/" +
            self.partStrList + "/Implement/freq" + str(freqID))
        refInputData = refInputData.replace("./dcpDir","./runs/" +
            self.partStrList + "/Checkpoint/freq" + str(freqID))
        refInputData = refInputData.replace("${top}_flpn.xdc", "freq" +
            str(freqID) + ".xdc")
        refInputData = refInputData.replace("set top \"uut_top\"",
            "set top \"" + self.topStr + "\"")
        refInputData = refInputData.replace("device", self.partStrList)
        #refInputData = refInputData.replace("devicePKG", self.partStrList[1])
        #refInputData = refInputData.replace("deviceSPEED", self.partStrList[2])
        #refInputData = refInputData.replace("deviceTEMP", self.partStrList[3])
        if fileName == self.vBR.runFilename:
            refInputData = refInputData.replace("#stratDef", optString)
        with open(os.path.abspath(self.prjLoc + "/result_repl/" +
            newGenFileName), 'w') as genFileOut:
            genFileOut.write(refInputData)

    def createVivadoFiles(self, fileName, start, n):
        #create gen/run TCL files for execution of runs
        with open(os.path.abspath(self.vBR.vivadoBatchLoc + "/scripts/" +
            fileName + ".tcl"), 'r') as genFileIn:
            genFileData = genFileIn.read()
        for i in range(start,n): self.createTCLFile(fileName, genFileData, i)

    def createPRJFile(self):
        #create PRJ files based on source list
        with open(os.path.abspath(self.prjLoc + "/" +
            "src_rtl/source_list.txt"), "r") as readFile:
            with open(os.path.abspath(self.prjLoc + "/" +
                "temp.prj"), "w") as writeFile:
                line = readFile.readline()
                prevLine = line
                while line:
                    if line and line != '\n' and line[0] != "#":
                        newLine = "vhdl work \"src_rtl/" + line.replace("\n","") + "\""
                        writeFile.write(newLine + "\n")
                    prevLine = line
                    line = readFile.readline()
        if self.top_name <0:
            self.topStr = prevLine.split('.')[0]
        else:
            self.topStr = self.top_name

    def createPRJFile_ResRep(self):
        #create PRJ files based on source list
        with open(os.path.abspath(self.prjLoc + "/" +
            "src_rtl/source_list.txt"), "r") as readFile:
            with open(os.path.abspath(self.prjLoc + "/result_repl/" +
                "temp.prj"), "w") as writeFile:
                line = readFile.readline()
                prevLine = line
                while line:
                    if line and line != '\n' and line[0] != "#":
                        newLine = "vhdl work \"src_rtl/" + line.replace("\n","") + "\""
                        writeFile.write(newLine + "\n")
                    prevLine = line
                    line = readFile.readline()

    def delAllVivadoRunsFiles(self):
        shutil.rmtree(self.prjLoc + "/runs/")

    def run16FreqList(self, n, f0, fn, WNSValList, algName):
        print algName
        WNSValList = [-100 for j in range(n)]
        threadList = []
        rerunList = []
        freqList = []
        periodList = []
        self.createClockFiles(n, f0, fn, freqList, periodList)
        print freqList
        print periodList
        for j in range(n): threadList.append(Thread(target=self.runGen, args=(j,)))
        for k in range(8):
            for j in range(16):
                time.sleep(10)
                threadList[j+(k*16)].start()
            time.sleep(10)
            for j in range(16): threadList[j+(k*16)].join()
            print "16runs num %d : Done" %(k+1)
            sys.stdout.flush()

        while True :
            i = 0
            del threadList[:]
            del rerunList[:]
            for j in range(n):
                try:
                    WNSValList[j] = self.getWNS(j,1)
                except:
                    print "error freq num : %d" %j
                    threadList.append(Thread(target=self.runGen, args=(j,)))
                    rerunList.append(j)
                    i= i+1
            print rerunList
            sys.stdout.flush()
            if i>0 :
                for k in range(i):
                    time.sleep(10)
                    threadList[k].start()
                for k in range(i): threadList[k].join()
            if i == 0:
                print "No Error"
                break

        print self.prjLoc + " : " + algName
        print freqList
        for j in range(n): print freqList[j]
        print "WNS Result :"
        print WNSValList
        for j in range(n): print WNSValList[j]
        sys.stdout.flush()
        return freqList

    def RunMinervaTPOptFreqList(self, max_NumRuns, RangeSize, f0, fn, algName,
        freqList, FreqAreaOptList, ReportFile):
        print algName
        WNSValList = [-100 for j in range(RangeSize)]
        threadList = []
        rerunList = []
        runList = []
        periodList = []
        del threadList[:]
        del freqList[:]
        del periodList[:]

        ReportFile.write("\n\n --- RunMinervaTPOptFreqList --- \n")
        print "Start RunMinervaTPOptFreqList f0= %d *** fn = %d *** RangeSize= %d" %(f0, fn, RangeSize)
        sys.stdout.flush()

        ReportFile.write("Inside RunMinervaTPOptFreqList f0= %d and fn = %d and RangeSize= %d \n" %(f0, fn, RangeSize))
        self.createClockFiles(RangeSize, f0, fn, freqList, periodList)
        ReportFile.write("Inside RunMinervaTPOptFreqList Frequency List: \n" +
            str(freqList) + "\n Inside RunMinervaTPOptFreqList Clock Period List: \n" +
            str(periodList) + "\n")

        NumRuns = RangeSize
        f_idx =0
        L_bound = 0
        U_bound = 0

        while True:

            if (NumRuns > max_NumRuns):
                NumThreads = max_NumRuns
            else:
                NumThreads = NumRuns

            NumThreads_Added = 0
            L_bound = U_bound
            U_bound = U_bound + NumThreads

            while True:
                if NumThreads_Added == NumThreads:
                    break

                threadList.append(Thread(target=self.runGen, args=(f_idx,)))
                runList.append(f_idx)
                f_idx += 1
                NumThreads_Added += 1

            for i in range(NumThreads):
                time.sleep(10)
                threadList[i].start()
            time.sleep(10)
            for i in range(NumThreads): threadList[i].join()
            ReportFile.write("%d runs: Done" %NumThreads + "\n")
            time.sleep(10)

            while True :
                i = 0
                del threadList[:]
                del rerunList[:]
                for j in range(L_bound, U_bound):
                    try:
                        WNSValList[j] = self.getWNS(j,1)
                        if WNSValList[j] > 0 :
                            FreqAreaOptList.append([freqList[j], self.getLUT(j,1), self.getSlice(j,1), self.getFF(j,1), 1])
                    except:
                        ReportFile.write("error freq num : %d" %j + "\n")
                        threadList.append(Thread(target=self.runGen, args=(j,)))
                        rerunList.append(j)
                        i = i+1
                ReportFile.write(str(rerunList) + "\n")
                if i>0 :
                    for k in range(i):
                        time.sleep(10)
                        threadList[k].start()
                    for k in range(i): threadList[k].join()
                if i == 0:
                    ReportFile.write("No Error \n")
                    break
            ReportFile.write("RunList: \n" + str(runList) + "\n")
            NumRuns = NumRuns - max_NumRuns
            if (NumRuns <= 0):
                break
        ReportFile.write(self.prjLoc + " : " + algName + "\n WNS Result : \n" + str(WNSValList) + "\n")
        return WNSValList

    def RunMinervaTPAOptFreqList(self, max_NumRuns, RangeSize, f0, fn, algName,
        freqList, FreqAreaOptList, ReportFile):
        print algName
        WNSValList = [-100 for j in range(RangeSize)]
        threadList = []
        rerunList = []
        runList = []
        periodList = []
        del threadList[:]
        del freqList[:]
        del periodList[:]

        ReportFile.write("\n\n --- RunMinervaTPAOptFreqList --- \n")
        print "Inside RunMinervaTPAOptFreqList f0= %d and fn = %d and RangeSize= %d" %(f0, fn, RangeSize)
        ReportFile.write("Inside RunMinervaTPAOptFreqList f0= %d and fn = %d and RangeSize= %d \n" %(f0, fn, RangeSize))
        self.createClockFiles_1MhzStep(f0, fn, freqList, periodList)
        ReportFile.write("Inside RunMinervaTPOptFreqList Frequency List: \n" +
            str(freqList) + "\n Inside RunMinervaTPOptFreqList Clock Period List: \n" +
            str(periodList) + "\n")

        NumRuns = RangeSize
        f_idx =0
        L_bound = 0
        U_bound = 0

        while True:
            if (NumRuns > max_NumRuns):
                NumThreads = max_NumRuns
            else:
                NumThreads = NumRuns

            NumThreads_Added = 0
            L_bound = U_bound
            U_bound = U_bound + NumThreads

            while True:
                if NumThreads_Added == NumThreads:
                    break

                threadList.append(Thread(target=self.runGen, args=(f_idx,)))
                runList.append(f_idx)
                f_idx += 1
                NumThreads_Added += 1

            for i in range(NumThreads):
                time.sleep(10)
                threadList[i].start()
            time.sleep(10)
            for i in range(NumThreads): threadList[i].join()
            ReportFile.write("%d runs: Done" %NumThreads + "\n")
            while True :
                i = 0
                del threadList[:]
                del rerunList[:]
                for j in range(L_bound, U_bound):
                    try:
                        WNSValList[j] = self.getWNS(j,1)
                        if WNSValList[j] > 0 :
                            FreqAreaOptList.append([freqList[j], self.getLUT(j,1), self.getSlice(j,1), self.getFF(j,1), 1])
                    except:
                        ReportFile.write("error freq num : %d" %j + "\n")
                        threadList.append(Thread(target=self.runGen, args=(j,)))
                        rerunList.append(j)
                        i = i+1
                ReportFile.write(str(rerunList) + "\n")
                if i>0 :
                    for k in range(i):
                        time.sleep(10)
                        threadList[k].start()
                    for k in range(i): threadList[k].join()
                if i == 0:
                    ReportFile.write("No Error \n")
                    break
            ReportFile.write("RunList: \n" + str(runList) + "\n")
            NumRuns = NumRuns - max_NumRuns
            if (NumRuns <= 0):
                break
        ReportFile.write(self.prjLoc + " : " + algName + "\n WNS Result : \n" + str(WNSValList) + "\n")
        return WNSValList

    def graphgenFreqSearch(self, f_ref, n, bestFreqIdxList, freqIdx):
        algName = "graphgenFreqSearch"
        WNSValList = []
        #freqList = []
        self.createPRJFile()
        self.createVivadoFiles(self.vBR.genFilename, 0, n)
        self.createVivadoFiles(self.vBR.runFilename, 0, n)

        f0 = f_ref - (n/2)
        fn = f_ref + (n/2)
        freqList = self.run16FreqList(n, f0, fn, WNSValList, algName)
        print freqList

        bestFreqIdxList[freqIdx] = 0

    def Minerva_TP_Opt(self, f0_in, fn_in, RangeSize_in, max_NumRuns,
        MaxFreqList, freqIdx, NumOpts, optList_2D, algName):
        alg_name = algName
        RangeSize = RangeSize_in
        h, w   = RangeSize, NumOpts+1
        WNSValList_1D = []
        WNSValList_2D = [[0 for x in range(w)] for y in range(h)]
        runList = []
        FreqAreaOptList = []
        FreqOverLUTList = []
        freqList = []
        optfreqList = []
        optperiodList = []
        threadListopt = []
        rerunList = []
        f0 = f0_in
        fn = fn_in
        firstRun_flag = 1
        lastRun_flag  = 0
        start = datetime.datetime.now()

        start_timestr = time.strftime("%Y-%m-%d_%H:%M")

        self.createPRJFile()
        self.createVivadoFiles(self.vBR.genFilename, 0, RangeSize)
        self.createVivadoFiles(self.vBR.runFilename, 0, RangeSize)

        result_elm = self.statusFileGen(alg_name)
        MTPOpt_elm = result_elm.find(alg_name)
        out_report = open (os.path.abspath(self.report_loc + "/" + 
            self.design_name + "_ID" + result_elm.get('id') + ".txt"), 'a', 1)
        out_report.write("======================= " + alg_name + " Routine =========================== \n\n")
        out_report.write("************************************************************** \n")
        out_report.write("***************** Minerva_TP_Opt ***************************** \n")
        out_report.write("************************************************************** \n \n \n")
        out_report.write("Device = " + str(self.partStrList) + "\n")
        out_report.write("Starting Time : \n")
        out_report.write(str(start) + "\n")

        while True:
            del freqList[:]
            del WNSValList_1D[:]
            WNSValList_2D = [[0 for x in range(w)] for y in range(h)]
            if firstRun_flag:
                out_report.write("Finding starting point... \n")
            while True:
                self.delAllVivadoRunsFiles()
                out_report.write("f0 = %d and fn = %d  ***** Before WNSValList_1D Call **** \n" %(f0, fn))
                WNSValList_1D = self.RunMinervaTPOptFreqList(max_NumRuns, RangeSize, f0, fn,
                alg_name, freqList, FreqAreaOptList, out_report)
                out_report.write("Frequency List : \n" + str(freqList) + "\n")
                posWNSIdx = -1
                for i in range(RangeSize):
                    if WNSValList_1D[i] >= 0:
                        posWNSIdx = i
                if (posWNSIdx == RangeSize-1):
                    if firstRun_flag:
                        f0 = fn
                        fn = fn + 100
                        out_report.write("step forward... \n")
                    else:
                        best_freq = freqList[posWNSIdx]
                        f0 = fn
                        fn = fn + RangeSize
                        out_report.write("**** fn has the best frequency before optimization runs **** \n == best frequency = %d == \n step forward... \n" %best_freq)
                        RunTime = (datetime.datetime.now()) - start
                        MTPOpt_elm.set('run_time', (str(RunTime))[:-10])
                        MTPOpt_elm.set('freq', str(best_freq))
                        MTPOpt_elm.set('LUT', str(self.getLUT(posWNSIdx,1)))
                        MTPOpt_elm.set('Slice', str(self.getSlice(posWNSIdx,1)))
                        MTPOpt_elm.set('FF', str(self.getFF(posWNSIdx,1)))
                        MTPOpt_elm.set('opt_id', '1')
                        with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                            + self.design_name + "_MS.xml"), 'w') as status_xml:
                            status_xml.write(self.prettify(self.ms_elm))
                elif (posWNSIdx < 0) :
                    if firstRun_flag:
                        fn = f0
                        f0 = f0 - 100
                        if (f0 < 0):
                            f0 = 1
                        out_report.write("step back... \n")
                    else:
                        out_report.write("start optimization rounds \n")
                        break
                else:
                    best_freq = freqList[posWNSIdx]
                    print "positive WNS is found for freq = %d  **** before optimization runs ******" %freqList[posWNSIdx]
                    sys.stdout.flush()
                    out_report.write("positive WNS is found for freq = %d  **** before optimization runs ****** \n" %freqList[posWNSIdx])
                    RunTime = (datetime.datetime.now()) - start
                    MTPOpt_elm.set('run_time', (str(RunTime))[:-10])
                    MTPOpt_elm.set('freq', str(best_freq))
                    MTPOpt_elm.set('LUT', str(self.getLUT(posWNSIdx,1)))
                    MTPOpt_elm.set('Slice', str(self.getSlice(posWNSIdx,1)))
                    MTPOpt_elm.set('FF', str(self.getFF(posWNSIdx,1)))
                    MTPOpt_elm.set('opt_id', '1')
                    MTPOpt_elm.set('tool_version', self.getVersion())
                    with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                        + self.design_name + "_MS.xml"), 'w') as status_xml:
                        status_xml.write(self.prettify(self.ms_elm))
                    break
            for i in range(RangeSize): WNSValList_2D[i][1] = WNSValList_1D[i]

            count = 0
            f_idx = posWNSIdx+1
            o_idx = 2
            allOpt_Done = 0
            while True:
                del runList[:]
                num_Runs = 0

                while True:
                    if f_idx == RangeSize:
                        f_idx = posWNSIdx+1
                        o_idx += 1
                    if (o_idx == NumOpts+1):
                        allOpt_Done = 1
                        break
                    if (num_Runs == max_NumRuns):
                        break
                    threadListopt.append(Thread(target=self.run,
                        args=(f_idx,o_idx, optList_2D[f_idx][o_idx])))
                    runList.append([f_idx,o_idx])
                    f_idx += 1
                    num_Runs += 1

                for j in range(num_Runs):
                    time.sleep(10)
                    threadListopt[j].start()
                time.sleep(10)
                for j in range(num_Runs): threadListopt[j].join()
                out_report.write("\nopt runs: Done \ncheck for missed runs... \n")
                while True :
                    m = 0
                    del threadListopt[:]
                    del rerunList[:]
                    for k in range(num_Runs):
                        f_idx_2 = runList[k][0]
                        o_idx_2 = runList[k][1]
                        o_id = self.getOptID(optList_2D[f_idx_2][o_idx_2])
                        try:
                            WNSValList_2D[f_idx_2][o_idx_2] = self.getWNS(f_idx_2, o_id)
                            if WNSValList_2D[f_idx_2][o_idx_2] > 0 :
                                FreqAreaOptList.append([freqList[f_idx_2],
                                self.getLUT(f_idx_2, o_id),
                                self.getSlice(f_idx_2, o_id),
                                self.getFF(f_idx_2, o_id), o_id])
                        except:
                            out_report.write("error freq num : %d,%d \n" %(f_idx_2, o_idx_2))
                            threadListopt.append(Thread(target=self.run,
                                args=(f_idx_2, o_idx_2, optList_2D[f_idx_2][o_idx_2])))
                            rerunList.append(f_idx_2)
                            m += 1
                    out_report.write(str(rerunList) + "\n")
                    if m>0 :
                        for k in range(m):
                            time.sleep(10)
                            threadListopt[k].start()
                        for k in range(m): threadListopt[k].join()
                    if m == 0:
                        out_report.write("No Error \n")
                        break
                print "All runs completed"
                print "Run List:"
                print runList
                out_report.write("All runs completed \n Run List: \n" +
                    str(runList) + "\n WNSValList_2D: \n" + str(WNSValList_2D) +
                    "\n" + str(FreqAreaOptList) + "\n")

                best_freq_idx = len(FreqAreaOptList)-1
                RunTime = (datetime.datetime.now()) - start
                MTPOpt_elm.set('run_time', (str(RunTime))[:-10])
                MTPOpt_elm.set('freq', str(FreqAreaOptList[best_freq_idx][0]))
                MTPOpt_elm.set('LUT', str(FreqAreaOptList[best_freq_idx][1]))
                MTPOpt_elm.set('Slice', str(FreqAreaOptList[best_freq_idx][2]))
                MTPOpt_elm.set('FF', str(FreqAreaOptList[best_freq_idx][3]))
                MTPOpt_elm.set('opt_id', str(FreqAreaOptList[best_freq_idx][4]))
                with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                    + self.design_name + "_MS.xml"), 'w') as status_xml:
                    status_xml.write(self.prettify(self.ms_elm))

                posWNSIdx_old = posWNSIdx
                posWNSIdx_tmp = -10
                for k in range(num_Runs):
                    if WNSValList_2D [runList[k][0]][runList[k][1]] > 0:
                        posWNSIdx_tmp = runList[k][0]
                if (posWNSIdx_tmp >= 0) and (posWNSIdx_tmp < RangeSize-1):
                    posWNSIdx = posWNSIdx_tmp
                    if (posWNSIdx+1) > f_idx:
                        f_idx = posWNSIdx+1
                    print "posWNSIdx updated"
                    print "New freq = %d" %freqList[posWNSIdx]
                    out_report.write("posWNSIdx updated \n New freq = %d \n" %freqList[posWNSIdx])
                elif posWNSIdx_tmp == (RangeSize-1):
                    best_freq = freqList[RangeSize-1]
                    posWNSIdx = posWNSIdx_tmp
                    f0 = freqList[RangeSize-1] + 1
                    RangeSize = RangeSize_in
                    fn = f0 + RangeSize
                    print "*** fn has positive WNS ***"
                    print "f0 = %d" %f0
                    print "== best_freq = %d ==" %best_freq
                    out_report.write("*** fn has positive WNS *** \n f0 = %d \n == best_freq = %d == \n" %(f0, best_freq))
                    break
                if allOpt_Done:
                    count += 1
                    out_report.write("*** opt run %d Done **** from freq-opt %s to freq-opt %s *** in %d to %d Freq Range **** \n"
                        %(count, runList[0], runList[num_Runs-1], freqList[posWNSIdx_old+1], fn))
                    if posWNSIdx < 0 :
                        lastRun_flag = 1
                        break
                    else:
                        if firstRun_flag:
                            f0 = int(freqList[posWNSIdx]) + 1
                            fn = f0 + RangeSize
                        else:
                            f0 = freqList[RangeSize-1] + 1
                            RangeSize = RangeSize_in - (RangeSize - (posWNSIdx+1))
                            fn = f0 + RangeSize
                        best_freq = freqList[posWNSIdx]
                        out_report.write("****** all optimizations run Done ****** \n == best_freq = %d == \n f0 = %d \n" %(best_freq, f0))
                        break
                count += 1
                print "*** opt run %d Done **** from freq-opt %s to freq-opt %s *** in %d to %d Freq Range ****" %(count, runList[0], runList[num_Runs-1], freqList[posWNSIdx_old+1], fn)
                out_report.write("*** opt run %d Done **** from freq-opt %s to freq-opt %s *** in %d to %d Freq Range **** \n"
                %(count, runList[0], runList[num_Runs-1], freqList[posWNSIdx_old+1], fn))
                sys.stdout.flush()

            firstRun_flag = 0
            if lastRun_flag:
                for m in range (len(FreqAreaOptList)):
                    FreqOverLUTList.append(float(float(FreqAreaOptList[m][0])/float(FreqAreaOptList[m][1])))
                MaxFreqoverLUT = FreqOverLUTList.index(max(FreqOverLUTList))
                Max_freq_idx = (len(FreqAreaOptList))-1
                out_report.write("**** "+ alg_name +" DONE **** \n")
                out_report.write("////////////////// Report Summary Start ////////////////// \n\n")
                out_report.write("Device = " + str(self.partStrList) + "\n")
                out_report.write("Top Name = " + self.topStr + "\n")
                out_report.write("Max Frequency: \n")
                out_report.write("Maximum Frequency = %d \nLuts = %d \nSlice = %d \nFF = %d \nOptimization ID = %d \n"
                    %(FreqAreaOptList[Max_freq_idx][0],
                    FreqAreaOptList[Max_freq_idx][1],
                    FreqAreaOptList[Max_freq_idx][2],
                    FreqAreaOptList[Max_freq_idx][3],
                    FreqAreaOptList[Max_freq_idx][4]))
                out_report.write("Max Frequency/LUT: \n")
                out_report.write("Frequency = %d \nLuts = %d \nSlice = %d \nFF = %d \nOptimization ID = %d \nFreq/LUT = %d \n"
                    %(FreqAreaOptList[MaxFreqoverLUT][0],
                    FreqAreaOptList[MaxFreqoverLUT][1],
                    FreqAreaOptList[MaxFreqoverLUT][2],
                    FreqAreaOptList[MaxFreqoverLUT][3],
                    FreqAreaOptList[MaxFreqoverLUT][4],
                    FreqOverLUTList[MaxFreqoverLUT]))
                break

        end = datetime.datetime.now()
        RunTime = end - start
        MTPOpt_elm.set('run_time', (str(RunTime))[:-10])
        MTPOpt_elm.set('freq', str(FreqAreaOptList[Max_freq_idx][0]))
        MTPOpt_elm.set('LUT', str(FreqAreaOptList[Max_freq_idx][1]))
        MTPOpt_elm.set('Slice', str(FreqAreaOptList[Max_freq_idx][2]))
        MTPOpt_elm.set('FF', str(FreqAreaOptList[Max_freq_idx][3]))
        MTPOpt_elm.set('opt_id', str(FreqAreaOptList[Max_freq_idx][4]))
        result_elm.set('is_complete', '1')
        with open (os.path.abspath(self.prjLoc + "/minerva_status/"
            + self.design_name + "_MS.xml"), 'w') as status_xml:
            status_xml.write(self.prettify(self.ms_elm))
        MaxFreqList[freqIdx] = best_freq
        out_report.write("Start Time = " + str(start) + "\nEnd Time = " +
            str(end) + "\nRun Time = %s" %str(RunTime))
        out_report.write("\n\n////////////////// Report Summary End ////////////////// \n\n")
        out_report.close()
        print "***************************************** "+ alg_name + " DONE *****************************************"

    def Minerva_TPA_Opt(self, best_freq, MaxNumOpt, optList_2D, MaxNumRuns,
        MaxFreqAreaList, BestAreaList, freqAreaIdx):
        algName = "Minerva_TPA_Opt"
        NumFreq = 0
        NumRuns = 0
        NumThreads = 0
        f_idx = 0
        o_idx = 0
        NumThreads_Added = 0
        OptBase_idx = 2
        freqList = []
        freqList_v2 = []
        runList = []
        threadListAreaOpt = []
        rerunList = []
        FreqAreaOptList = []
        FreqOverLUTList = []
        FreqOverLUTList_temp = []
        L_bound = int(0.96 * best_freq)
        U_bound = int(best_freq)

        for i in range(L_bound, U_bound+1):
            freqList.append(i)
            NumFreq += 1
        h, w   = NumFreq, MaxNumOpt
        WNSValList_2D = [[0 for x in range(w)] for y in range(h)]
        start = datetime.datetime.now()

        self.createPRJFile()
        self.createVivadoFiles(self.vBR.genFilename, 0, NumFreq)
        self.createVivadoFiles(self.vBR.runFilename, 0, NumFreq)
        self.delAllVivadoRunsFiles()
        result_elm = self.statusFileGen(algName)
        MTPAOpt_elm = result_elm.find(algName)

        out_report = open (os.path.abspath(self.report_loc + "/" + 
            self.design_name + "_ID" + result_elm.get('id') + ".txt"), 'a', 1)
        out_report.write("************************************************************** \n")
        out_report.write("******************** Minerva_TPA_Opt *************************** \n")
        out_report.write("************************************************************** \n \n \n")
        out_report.write("Start Minerva_TPA_Opt from %d MHz to %d MHz ... \n"
            %(L_bound, U_bound))
        out_report.write("Device = " + str(self.partStrList) + "\n")
        out_report.write(str(freqList) + "\n")
        out_report.write("Starting Time : \n")
        out_report.write(str(start) + "\n")

        WNSValList_1D = self.RunMinervaTPAOptFreqList(MaxNumRuns, NumFreq,
            L_bound, U_bound, algName, freqList_v2, FreqAreaOptList, out_report)
        for i in range(NumFreq): WNSValList_2D[i][1] = WNSValList_1D[i]
        NumRuns = NumFreq * (MaxNumOpt - OptBase_idx)
        o_idx = OptBase_idx

        while(True):
            NumThreads = 0
            NumThreads_Added = 0
            if (NumRuns > MaxNumRuns):
                NumThreads = MaxNumRuns
            else:
                NumThreads = NumRuns

            del threadListAreaOpt [:]
            del runList [:]
            while True:

                if (o_idx == MaxNumOpt):
                    f_idx +=1
                    o_idx = OptBase_idx

                if f_idx == NumFreq:
                    break

                if NumThreads_Added == NumThreads:
                    break

                threadListAreaOpt.append(Thread(target=self.run,
                    args=(f_idx, o_idx,optList_2D[f_idx][o_idx])))
                runList.append([f_idx,o_idx])
                o_idx += 1
                NumThreads_Added += 1
            out_report.write("NumRuns = %d \n" %NumRuns)
            out_report.write("NumThreads_Added = %d \n" %NumThreads_Added)
            for i in range(NumThreads_Added):
                time.sleep(10)
                threadListAreaOpt[i].start()
            time.sleep(10)
            for i in range(NumThreads_Added): threadListAreaOpt[i].join()
            time.sleep(10)
            out_report.write("opt runs: Done \n")
            out_report.write("check for missed runs... \n")
            while True :
                m = 0
                del threadListAreaOpt[:]
                del rerunList[:]
                for k in range(NumThreads_Added):
                    f_idx_2 = runList[k][0]
                    o_idx_2 = runList[k][1]
                    o_id = self.getOptID(optList_2D[f_idx_2][o_idx_2])
                    try:
                        WNSValList_2D[f_idx_2][o_idx_2] = self.getWNS(f_idx_2, o_id)
                        if WNSValList_2D[f_idx_2][o_idx_2] > 0 :
                            FreqAreaOptList.append([freqList[f_idx_2],
                            self.getLUT(f_idx_2, o_id), self.getSlice(f_idx_2, o_id),
                            self.getFF(f_idx_2, o_id), o_id])
                    except:
                        out_report.write("error freq num : %d,%d \n" %(f_idx_2, o_idx_2))
                        threadListAreaOpt.append(Thread(target=self.run,
                            args=(f_idx_2, o_idx_2, optList_2D[f_idx_2][o_idx_2])))
                        rerunList.append(f_idx_2)
                        m += 1
                out_report.write(str(rerunList) + "/n")
                if m>0 :
                    for k in range(m):
                        time.sleep(10)
                        threadListAreaOpt[k].start()
                    time.sleep(10)
                    for k in range(m): threadListAreaOpt[k].join()
                if m == 0:
                    # print "No Error"
                    out_report.write("No Error /n")
                    break
            out_report.write("All runs completed /n Run List: \n" + str(runList) +
                "\n WNSValList_2D: \n" + str(WNSValList_2D) +
                " \n FreqAreaOptList: \n" + str(FreqAreaOptList) + "\n")

            for m in range (len(FreqAreaOptList)):
                FreqOverLUTList.append(float(float(FreqAreaOptList[m][0])/float(FreqAreaOptList[m][1])))
            MaxFreqoverLUT = FreqOverLUTList.index(max(FreqOverLUTList))
            curr_time = datetime.datetime.now()
            RunTime = curr_time - start
            MTPAOpt_elm.set('run_time', (str(RunTime))[:-10])
            MTPAOpt_elm.set('freq', str(FreqAreaOptList[MaxFreqoverLUT][0]))
            MTPAOpt_elm.set('LUT', str(FreqAreaOptList[MaxFreqoverLUT][1]))
            MTPAOpt_elm.set('Slice', str(FreqAreaOptList[MaxFreqoverLUT][2]))
            MTPAOpt_elm.set('FF', str(FreqAreaOptList[MaxFreqoverLUT][3]))
            MTPAOpt_elm.set('opt_id', str(FreqAreaOptList[MaxFreqoverLUT][4]))
            MTPOpt_elm.set('tool_version', self.getVersion())
            with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                + self.design_name + "_MS.xml"), 'w') as status_xml:
                status_xml.write(self.prettify(self.ms_elm))

            NumRuns = NumRuns - MaxNumRuns
            if (NumRuns <= 0):
                result_elm.set('is_complete', '1')
                with open (os.path.abspath(self.prjLoc + "/minerva_status/"
                    + self.design_name + "_MS.xml"), 'w') as status_xml:
                    status_xml.write(self.prettify(self.ms_elm))
                out_report.write(str(FreqOverLUTList) + "/n")
                out_report.write("===================================/n" +
                    str(FreqAreaOptList[MaxFreqoverLUT]) + "\n=================================== /n")
                out_report.write("**** Area Optimization DONE **** \n")
                out_report.write("////////////////// Report Summary Start ////////////////// \n\n")
                out_report.write("Device = " + str(self.partStrList) + "\n")
                out_report.write("Top Name = " + self.topStr + "\n")
                out_report.write("Frequency = %d MHz \nLUTs = %d \nSlice = %d \nFF = %d \nOptimization ID = %d \nFreq/LUT = %d \n"
                    %(FreqAreaOptList[MaxFreqoverLUT][0],
                    FreqAreaOptList[MaxFreqoverLUT][1],
                    FreqAreaOptList[MaxFreqoverLUT][2],
                    FreqAreaOptList[MaxFreqoverLUT][3],
                    FreqAreaOptList[MaxFreqoverLUT][4], max(FreqOverLUTList)))

                end = datetime.datetime.now()
                RunTime = end - start
                out_report.write("Start Time = " + str(start) +
                    "\nEnd Time = " + str(end) + "\nRun Time = %s" %str(RunTime))
                out_report.write("\n\n////////////////// Report Summary End ////////////////// \n\n")
                MaxFreqAreaList[freqAreaIdx] = FreqAreaOptList[MaxFreqoverLUT][0]
                BestAreaList[freqAreaIdx] = FreqAreaOptList[MaxFreqoverLUT][1]
                break
        out_report.close()

    def BinaryFreqSearch(self, f0_in, fn_in, runID, MaxFreqList):

        print "ID %d: Start binary search..." %runID
        now = datetime.datetime.now()
        print "ID %d: Starting Time :" %runID
        print str(now)

        threadList = []
        f0 = f0_in
        fn = fn_in

        self.createPRJFile()
        self.createVivadoFiles(self.vBR.genFilename, 0, 1)
        self.createVivadoFiles(self.vBR.runFilename, 0, 1)

        rnds = 0
        oldfreq = 0
        while True:

            currfreq = float((f0 + fn)/2)
            print "ID %d: round %d" %(runID, rnds)
            print "ID %d: f0 = %.2f" %(runID, f0)
            print "ID %d: fn = %.2f" %(runID, fn)
            print "ID %d: Freq = %.2f" %(runID, currfreq)

            dif = float(currfreq - oldfreq)
            if ((dif>0) and (dif<1)) or (rnds>25):
                MaxFreqList[runID] = oldfreq
                LUT = self.getLUT(0,1)
                print "              ********** ID %d: Max Freq = %d --- LUT = %d **********" %(runID, oldfreq, LUT)
                break

            self.delAllVivadoRunsFiles()
            currPeriod = float(1000/currfreq)
            self.createXDCClock(self.prjLoc + "/" +
                vivadoBatchRun.XDCFILEPREFIX + str(0) + ".xdc", currPeriod)
            threadList.append(Thread(target=self.runGen, args=(0,)))
            time.sleep(10)
            threadList[0].start()
            time.sleep(10)
            threadList[0].join()
            print "ID %d: Run Done" %runID

            while True:
                del threadList[:]
                try:
                    WNS = self.getWNS(0,1)
                    print "ID %d: WNS = %d" %(runID, WNS)
                    break
                except:
                    print "ID %d: Error occured" %runID
                    print "ID %d: Start rerun" %runID
                    threadList.append(Thread(target=self.runGen, args=(0,)))
                    time.sleep(10)
                    threadList[0].start()
                    time.sleep(10)
                    threadList[0].join()

            if WNS > 0:
                f0 = currfreq
            else:
                fn = currfreq

            oldfreq = currfreq
            rnds +=1
            sys.stdout.flush()

        now = datetime.datetime.now()
        print "ID %d: End Time :" %runID
        print str(now)

    def SingleRun(self, freq, o_idx, optString):
        algName = "Minerva_Single"
        start = datetime.datetime.now()

        self.createPRJFile()
        with open(os.path.abspath(self.vBR.vivadoBatchLoc + "/scripts/" +
            self.vBR.genFilename + ".tcl"), 'r') as genFileIn:
            genFileData = genFileIn.read()
        self.createTCLFile(self.vBR.genFilename, genFileData, freq)
        with open(os.path.abspath(self.vBR.vivadoBatchLoc + "/scripts/" +
            self.vBR.runFilename + ".tcl"), 'r') as genFileIn:
            genFileData = genFileIn.read()
        self.createTCLFile(self.vBR.runFilename, genFileData, freq)

        result_elm = self.statusFileGen(algName)
        Single_elm = result_elm.find(algName)

        out_report = open (os.path.abspath(self.report_loc + "/" + 
            self.design_name + "_ID" + result_elm.get('id') + ".txt"), 'a', 1)
        out_report.write("************************************************************** \n")
        out_report.write("***************** SINGLE RUN ********************************* \n")
        out_report.write("************************************************************** \n \n \n")

        self.delAllVivadoRunsFiles()
        Period = float(1000/float(freq))
        sys.stdout.flush()
        self.createXDCClock(self.prjLoc + "/" +
            vivadoBatchRun.XDCFILEPREFIX + str(freq) + ".xdc", Period)
        self.runGen(freq)
        self.run(freq, o_idx, optString)

        wns = self.getWNS(freq, o_idx)
        lut = self.getLUT(freq, o_idx)
        Slice = self.getSlice(freq, o_idx)
        ff = self.getFF(freq, o_idx)

        end = datetime.datetime.now()
        RunTime = end - start
        Single_elm.set('run_time', (str(RunTime))[:-10])
        Single_elm.set('freq', str(freq))
        Single_elm.set('LUT', lut)
        Single_elm.set('Slice', Slice)
        Single_elm.set('FF', ff)
        Single_elm.set('opt_id', str(o_idx))
        result_elm.set('is_complete', '1')
        out_report.write("////////////////// Report Summary Start ////////////////// \n\n")
        out_report.write("Device = " + str(self.partStrList) + "\n")
        out_report.write("Frequency = %d MHz \nLUTs = %d \nSlice = %d \nFF = %d \nOptimization ID = %d \nFreq/LUT = %.2f"
            %(freq, lut, Slice, ff, o_idx, float(float(freq)/float(lut))))
        out_report.write("\n\n////////////////// Report Summary End ////////////////// \n\n")
        out_report.close()

    def ResultReplicationPkgGen(self, freq, o_idx, optString, topStr,
        ToolVer, ResId, SrcMode, UrlText):

        self.topStr = topStr
        if not os.path.exists(os.path.abspath(self.prjLoc + "/result_repl")):
            os.makedirs(os.path.abspath(self.prjLoc + "/result_repl"))

        self.createPRJFile_ResRep()
        with open(os.path.abspath(self.vBR.vivadoBatchLoc + "/scripts/" +
            self.vBR.genFilename + ".tcl"), 'r') as genFileIn:
            genFileData = genFileIn.read()
        self.createTCLFile_ResRep(self.vBR.genFilename, genFileData, freq,
            o_idx, optString)
        with open(os.path.abspath(self.vBR.vivadoBatchLoc + "/scripts/" +
            self.vBR.runFilename + "_ResRep.tcl"), 'r') as genFileIn:
            genFileData = genFileIn.read()
        self.createTCLFile_ResRep(self.vBR.runFilename, genFileData, freq,
            o_idx, optString)

        Period = float(1000/float(freq))
        self.createXDCClock(self.prjLoc + "/result_repl/" +
            vivadoBatchRun.XDCFILEPREFIX + str(freq) + ".xdc", Period)

        run_bash = open (os.path.abspath((self.prjLoc + "/result_repl/run.sh")),
            'a', 1)
        run_bash.write("#!/bin/bash \n")
        run_bash.write("vivado -mode batch -source " + self.vBR.genFilename +
            "_freq" + str(freq) + "_" + str(o_idx) +".tcl -notrace > synth_out"+
            str(freq)+ "_" + str(o_idx) + ".txt\n")
        run_bash.write("vivado -mode batch -source " + self.vBR.runFilename +
            "_freq" + str(freq) + "_" + str(o_idx) +".tcl -notrace > impl_out" +
            str(freq)+ "_" + str(o_idx) + ".txt\n")
        run_bash.close()

        run_bat = open (os.path.abspath((self.prjLoc + "/result_repl/run.bat")),
            'a', 1)
        run_bat.write("vivado -mode batch -source " + self.vBR.genFilename +
            "_freq" + str(freq) + "_" + str(o_idx) +".tcl -notrace > synth_out"+
            str(freq)+ "_" + str(o_idx) + ".txt && vivado -mode batch -source "+
            self.vBR.runFilename + "_freq" + str(freq) + "_" + str(o_idx) +
            ".tcl -notrace > impl_out" + str(freq)+ "_" + str(o_idx) + ".txt\n")
        run_bat.close()

        shutil.copytree(os.path.abspath(self.vBR.vivadoBatchLoc +
            "/scripts/Tcl"),os.path.abspath(self.prjLoc + "/result_repl/Tcl"))

        if SrcMode == 1:
            shutil.copytree(os.path.abspath(self.prjLoc + "/src_rtl"),
                os.path.abspath(self.prjLoc + "/result_repl/src_rtl"))
        elif SrcMode == 2:
            os.makedirs(os.path.abspath(self.prjLoc + "/result_repl/src_rtl"))
            with open (os.path.abspath(self.prjLoc +
                "/result_repl/src_rtl/url.txt"), 'w') as url:
                url.write(UrlText)
        elif SrcMode == 3:
            os.makedirs(os.path.abspath(self.prjLoc + "/result_repl/src_rtl"))

        with open (os.path.abspath(self.vBR.vivadoBatchLoc +
            "/scripts/run_ResRep_temp.tcl"), 'r') as run_tcl_tmp:
            run_tcl_data = run_tcl_tmp.read()
            run_tcl_data = run_tcl_data.replace("*tool_version*", ToolVer)

        with open (os.path.abspath(self.prjLoc + "/result_repl/Tcl/run.tcl"),
            'w') as run_tcl:
            run_tcl.write(run_tcl_data)

        #! copy the readme file
        with open (os.path.abspath(self.vBR.vivadoBatchLoc +
            "/scripts/readme_temp.txt"), 'r') as readme_tmp:
            readme_data = readme_tmp.read()
            readme_data = readme_data.replace("*tool_version*", ToolVer)
        with open (os.path.abspath(self.prjLoc
            + "/result_repl/readme.txt"), 'w') as readme:
            readme.write(readme_data)

        #! copy status xml to the result replication package
        status_xml_path = os.path.abspath(self.prjLoc + "/minerva_status/" +
            self.design_name + "_MS.xml")
        with open (os.path.abspath(self.prjLoc + "/result_repl/result.xml")
            , 'w') as res_xml:
            tree = ET.parse(status_xml_path)
            ms_elm = tree.getroot()
            sl_elm = ms_elm.find('StatusList')
            for result_elm in sl_elm:
                if int(result_elm.get('id')) == ResId:
                    for st_elm in result_elm:
                        status_elm = st_elm
                    res_xml.write(ET.tostring(status_elm, 'utf-8'))

        timestr = time.strftime("%Y_%m_%d-%H_%M")
        os.chdir(self.prjLoc)
        shutil.make_archive(os.path.abspath(self.prjLoc + "/result_repl_" +
            self.design_name + "_" + timestr), "zip", self.prjLoc, "result_repl")
        shutil.rmtree(os.path.abspath(self.prjLoc + "/result_repl/"))

    def debug(self):
        print "test"
        string = "add_implementation TopDown8\n" + \
                    "set_attribute impl TopDown8      top          $top\n" + \
                    "set_attribute impl TopDown8      implXDC      [list $xdcDir/freq" + ".xdc]\n" + \
                    "set_attribute impl TopDown8      td.impl      0\n" + \
                    "set_attribute impl TopDown8      impl         ${run.tdImpl}\n" + \
                    "set_attribute impl TopDown8   place_directive       ExtraNetDelay_medium\n" + \
                    "set_attribute impl TopDown8   route_directive       AdvancedSkewModeling\n"
        print self.getVersion()
